# Default PW Instructions: Specific specifications of different fields used to
# import and export data to and from PowerWorld. The goal here is to avoid 
# duplicating work in reading and writing PowerWorld fields, and reduce bugs.
#
# Adam Birchfield, Texas A&M University
# 
# Log:
# 9/3/2022 Initial version, based on prior.
#

def make_default_pw_instructions(wb):

    wb.pw_instructions = {

        "region.name": {"pwfield":["SAName", "Name"],
            "import_from_aux": "string_main"},

        "area.number": {"pwfield":["AreaNum", "Number"],
            "import_from_aux": "int_main"}, 
        "area.name": {"pwfield":["AreaName", "Name"],
            "import_from_aux": "string"},
        "area.region": {"pwfield":["SAName"],
            "import_from_aux": "string",
            "export_from_aux": ".name"}, 

        "sub.number": {"pwfield":["SubNum", "Number"],
            "import_from_aux": "int_main"}, 
        "sub.name": {"pwfield":["SubName", "Name"],
            "import_from_aux": "string"},
        "sub.latitude": {"pwfield":["Latitude"],
            "import_from_aux": "float"},
        "sub.longitude": {"pwfield":["Longitude"],
            "import_from_aux": "float"},
        "sub.area": {"pwfield":["AreaNum"],
            "import_from_aux": "int",
            "export_from_aux": ".number"}, 

        "bus.number": {"pwfield":["BusNum", "Number"],
            "import_from_aux": "int_main"}, 
        "bus.name": {"pwfield":["BusName", "Name"],
            "import_from_aux": "string"},
        "bus.nominal_kv": {"pwfield":["BusNomVolt", "NomkV"],
            "import_from_aux": "float"},
        "bus.vpu": {"pwfield":["BusPUVolt", "Vpu"],
            "import_from_aux": "float"},
        "bus.vang": {"pwfield":["BusAngle", "Vangle"],
            "import_from_aux": "float"},
        "bus.status": {"pwfield":["BusStatus"], 
            "request_from_simauto":"connected",
            "import_from_aux": "connected",
            "export_from_aux": "connected"},
        "bus.area": {"pwfield":["AreaNum", "AreaNumber"],
            "import_from_aux": "int",
            "export_from_aux": ".number"}, 
        "bus.sub": {"pwfield":["SubNum", "SubNumber"],
            "import_from_aux": "int",
            "export_from_aux": ".number"}, 
        "bus.zone_number": {"pwfield":["ZoneNum"],
            "import_from_aux": "int"},

        "gen.bus": {"pwfield":["BusNum", "Number"],
            "import_from_aux": "int",
            "export_from_aux": ".number"}, 
        "gen.id": {"pwfield":["GenID", "ID"],
            "import_from_aux": "string"}, 
        "gen.status": {"pwfield":["GenStatus", "Status"], 
            "request_from_simauto":"closed",
            "import_from_aux": "closed",
            "export_from_aux": "closed"},
        "gen.p": {"pwfield":["GenMW", "GenMW"],
            "import_from_aux": "float"},
        "gen.q": {"pwfield":["GenMVR", "GenMVR"],
            "import_from_aux": "float"},
        "gen.pset": {"pwfield":["GenMWSetPoint", "MWSetPoint"],
            "import_from_aux": "float"},
        "gen.qset": {"pwfield":["GenMvrSetPoint", "MvarSetPoint"],
            "import_from_aux": "float"},
        "gen.pmax": {"pwfield":["GenMWMax", "MWMax"],
            "import_from_aux": "float"},
        "gen.pmin": {"pwfield":["GenMWMin", "MWMin"],
            "import_from_aux": "float"},
        "gen.qmax": {"pwfield":["GenMVRMax", "MvarMax"],
            "import_from_aux": "float"},
        "gen.qmin": {"pwfield":["GenMVRMin", "MvarMin"],
            "import_from_aux": "float"},
        "gen.reg_bus_num": {"pwfield":["GenRegNum", "RegBusNum"],
            "import_from_aux": "int"},
        "gen.reg_pu_v": {"pwfield":["GenVoltSet", "VoltSet"],
            "import_from_aux": "float"},
        "gen.p_auto_status": {"pwfield":["GenAGCAble", "AGC"], 
            "request_from_simauto":"yesno",
            "import_from_aux": "yesno",
            "export_from_aux": "yesno"},
        "gen.q_auto_status": {"pwfield":["GenAVRAble", "AVR"], 
            "request_from_simauto":"yesno",
            "import_from_aux": "yesno",
            "export_from_aux": "yesno"},
        "gen.ramp_rate": {"pwfield":["GenMWRampLimit"],
            "import_from_aux": "float"},
        "gen.crank_p": {"pwfield":["CustomFloat"],
            "import_from_aux": "float"},
        "gen.crank_t": {"pwfield":["CustomFloat:1"],
            "import_from_aux": "float"},
        "gen.availability": {"pwfield":["CustomFloat:2"],
            "import_from_aux": "float"},
        "gen.fuel_type": {"pwfield":["GenFuelType", "FuelType"],
            "import_from_aux": "string"},
        "gen.fuel_cost": {"pwfield":["GenFuelCost", "FuelCost"],
            "import_from_aux": "float"},
        "gen.sbase": {"pwfield":["GenMVABase", "MVABase"],
            "import_from_aux": "float"},
        "gen.fixed_cost": {"pwfield":["GenFixedCost", "FixedCost"],
            "import_from_aux": "float"},
        "gen.cost_b": {"pwfield":["GenIOB", "CubicB"],
            "import_from_aux": "float"},
        "gen.cost_c": {"pwfield":["GenIOC", "CubicC"],
            "import_from_aux": "float"},
        "gen.cost_d": {"pwfield":["GenIOD", "CubicD"],
            "import_from_aux": "float"},

        "load.bus": {"pwfield":["BusNum", "Number"],
            "import_from_aux": "int",
            "export_from_aux": ".number"}, 
        "load.id": {"pwfield":["LoadID", "ID"],
            "import_from_aux": "string"}, 
        "load.status": {"pwfield":["LoadStatus", "Status"], 
            "request_from_simauto":"closed",
            "import_from_aux": "closed",
            "export_from_aux": "closed"},
        "load.p": {"pwfield":["LoadMW", "MW"],
            "import_from_aux": "float"},
        "load.q": {"pwfield":["LoadMVR", "Mvar"],
            "import_from_aux": "float"},
        "load.ps": {"pwfield":["LoadSMW", "SMW"],
            "import_from_aux": "float"},
        "load.qs": {"pwfield":["LoadSMVR", "SMvar"],
            "import_from_aux": "float"},
        "load.benefit": {"pwfield":["GenBidMWHR"],
            "import_from_aux": "float"},

        "shunt.bus": {"pwfield":["BusNum", "Number"],
            "import_from_aux": "int",
            "export_from_aux": ".number"}, 
        "shunt.id": {"pwfield":["ShuntID", "ID"],
            "import_from_aux": "string"}, 
        "shunt.status": {"pwfield":["SSStatus", "Status"], 
            "request_from_simauto":"closed",
            "import_from_aux": "closed",
            "export_from_aux": "closed"},
        "shunt.qnom": {"pwfield":["SSNMVR", "NMVR", "MvarNom"],
            "import_from_aux": "float"},
        "shunt.q": {"pwfield":["SSAMVR", "MVR"],
            "import_from_aux": "float"},
        "shunt.availability": {"pwfield":["CustomInteger"],
            "import_from_aux": "int"},
        "shunt.control_mode": {"pwfield":["SSCMode", "ShuntMode"],
            "import_from_aux": "string"},

        "branch.from_bus": {"pwfield":["BusNum", "BusFrom", "BusNumFrom", ],
            "import_from_aux": "int",
            "export_from_aux": ".number"}, 
        "branch.to_bus": {"pwfield":["BusNum:1", "BusTo", "BusNumTo", ],
            "import_from_aux": "int",
            "export_from_aux": ".number"}, 
        "branch.id": {"pwfield":["LineCircuit", "Circuit"],
            "import_from_aux": "string"}, 
        "branch.status": {"pwfield":["LineStatus", "Status"], 
            "request_from_simauto":"closed",
            "import_from_aux": "closed",
            "export_from_aux": "closed"},
        "branch.R": {"pwfield":["LineR", "LineR:1", "R", "Rxfbase"],
            "import_from_aux": "float"},
        "branch.X": {"pwfield":["LineX", "LineX:1", "X", "Xxfbase"],
            "import_from_aux": "float"},
        "branch.B": {"pwfield":["LineC", "LineC:1", "B", "Bxfbase"],
            "import_from_aux": "float"},
        "branch.G": {"pwfield":["LineG", "LineG:1", "G", "Gxfbase"],
            "import_from_aux": "float"},
        "branch.MVA_Limit_A": {"pwfield":["LineAMVA"],
            "import_from_aux": "float"},
        "branch.MVA_Limit_B": {"pwfield":["LineAMVA:1"],
            "import_from_aux": "float"},
        "branch.MVA_Limit_C": {"pwfield":["LineAMVA:2"],
            "import_from_aux": "float"},
        "branch.p1": {"pwfield":["LineMW"],
            "import_from_aux": "float"},
        "branch.q1": {"pwfield":["LineMVR"],
            "import_from_aux": "float"},
        "branch.p2": {"pwfield":["LineMW:1"],
            "import_from_aux": "float"},
        "branch.q2": {"pwfield":["LineMVR:1"],
            "import_from_aux": "float"},
        "branch.length": {"pwfield":["GICLineDistance"],
            "import_from_aux": "float"},
        "branch.availability": {"pwfield":["CustomInteger"],
            "import_from_aux": "int"},
        "branch.branch_device_type": {"pwfield":["BranchDeviceType"],
            "import_from_aux": "string"},

    }